# Copyright 2021 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic types for reprenting unknown types and objects."""

from typing import Annotated, Any, ClassVar, Literal
from pyglove.core import typing as pg_typing
from pyglove.core import utils
from pyglove.core.symbolic import list as pg_list  # pylint: disable=unused-import
from pyglove.core.symbolic import object as pg_object


class UnknownSymbol(pg_object.Object, pg_typing.CustomTyping):
  """Interface for symbolic representation of unknown symbols."""
  auto_register = False

  def custom_apply(self, *args, **kwargs) -> tuple[bool, Any]:
    """Bypass PyGlove type check."""
    return (False, self)


class UnknownType(UnknownSymbol):
  """Symbolic object for representing unknown types."""

  auto_register = True
  __serialization_key__ = 'unknown_type'

  # TODO(daiyip): Revisit the design on how `pg.typing.Object()` handles
  # UnknownType. This hacky solution should be removed in the future.
  __no_type_check__ = True

  name: str
  args: list[Any] = []

  def sym_jsonify(self, **kwargs) -> utils.JSONValueType:
    json_dict = {'_type': 'type', 'name': self.name}
    if self.args:
      json_dict['args'] = utils.to_json(self.args, **kwargs)
    return json_dict

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      **kwargs
  ) -> str:
    s = f'<unknown-type {self.name}>'
    if self.args:
      s += f'[{", ".join(repr(x) for x in self.args)}]'
    return s

  def __call__(self, **kwargs):
    return UnknownTypedObject(
        type_name=self.name, **kwargs
    )


class UnknownCallable(UnknownSymbol):
  """Symbolic object for representing unknown callables."""

  auto_register = False
  name: str
  CALLABLE_TYPE: ClassVar[Literal['function', 'method']]

  def sym_jsonify(self, **kwargs) -> utils.JSONValueType:
    return {'_type': self.CALLABLE_TYPE, 'name': self.name}

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      **kwargs
  ) -> str:
    return f'<unknown-{self.CALLABLE_TYPE} {self.name}>'


class UnknownFunction(UnknownCallable):
  """Symbolic objject for representing unknown functions."""

  auto_register = True
  __serialization_key__ = 'unknown_function'
  CALLABLE_TYPE = 'function'


class UnknownMethod(UnknownCallable):
  """Symbolic object for representing unknown methods."""

  auto_register = True
  __serialization_key__ = 'unknown_method'
  CALLABLE_TYPE = 'method'


class UnknownTypedObject(UnknownSymbol):
  """Symbolic object for representing objects of unknown-type."""

  auto_register = True
  __serialization_key__ = 'unknown_object'

  type_name: str
  __kwargs__: Annotated[
      Any,
      (
          'Fields of the original object will be kept as symbolic attributes '
          'of this object so they can be accessed through `__getattr__`.'
      )
  ]

  def sym_jsonify(self, **kwargs) -> utils.JSONValueType:
    """Converts current object to a dict of plain Python objects."""
    json_dict = self._sym_attributes.to_json(
        exclude_keys=set(['type_name']), **kwargs
    )
    assert isinstance(json_dict, dict)
    json_dict[utils.JSONConvertible.TYPE_NAME_KEY] = self.type_name
    return json_dict

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      **kwargs
  ) -> str:
    exclude_keys = kwargs.pop('exclude_keys', set())
    exclude_keys.add('type_name')
    kwargs['exclude_keys'] = exclude_keys
    return self._sym_attributes.format(
        compact,
        verbose,
        root_indent,
        cls_name=f'<unknown-type {self.type_name}>',
        key_as_attribute=True,
        bracket_type=utils.BracketType.ROUND,
        **kwargs,
    )



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from pyglove.core import symbolic   # pylint: disable=unused-import
from pyglove.core.views.html import base
from pyglove.core.views.html.controls import tooltip as tooltip_lib


class TooltipTest(unittest.TestCase):

  def assert_html_content(self, control, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = control.to_html().content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_basic(self):
    tooltip = tooltip_lib.Tooltip('foo')
    with self.assertRaisesRegex(
        ValueError, 'CSS selector `for_element` is required'
    ):
      tooltip.to_html()

    tooltip = tooltip_lib.Tooltip('foo', for_element='.bar')
    self.assertEqual(tooltip.for_element, '.bar')
    self.assert_html_content(
        tooltip,
        '<span class="tooltip">foo</span>'
    )
    self.assertIn(
        inspect.cleandoc(
            """
            .bar:hover + .tooltip {
              visibility: visible;
            }
            """
        ),
        tooltip.to_html().style_section,
    )

  def test_update(self):
    tooltip = tooltip_lib.Tooltip('foo', for_element='.bar', interactive=True)
    self.assertIn('id="control-', tooltip.to_html_str(content_only=True))
    with tooltip.track_scripts() as scripts:
      tooltip.update('normal text')
      self.assertEqual(tooltip.content, 'normal text')
      self.assertEqual(
          scripts,
          [
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{tooltip.element_id()}");
                  elem.textContent = "normal text";
                  """
              ),
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{tooltip.element_id()}");
                  elem.classList.remove("html-content");
                  """
              ),
          ]
      )
    with tooltip.track_scripts() as scripts:
      tooltip.update(base.Html('<b>bold text</b>'))
      self.assertEqual(
          scripts,
          [
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{tooltip.element_id()}");
                  elem.innerHTML = "<b>bold text</b>";
                  """
              ),
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{tooltip.element_id()}");
                  elem.classList.add("html-content");
                  """
              ),
          ]
      )

if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Genotype for numerical decisions."""

import random
import types
from typing import Any, List, Optional, Union

from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core import utils
from pyglove.core.geno.base import DecisionPoint
from pyglove.core.geno.base import DNA


def float_scale_spec(field_name):
  """Returns value spec for the scale of a continuous range."""
  return (field_name, pg_typing.Enum(None, [None, 'linear', 'log', 'rlog']),
          ('The scale of values within the range for the search algorithm '
           'to explore. '
           'If None, the feasible space is unscaled;'
           'If `linear`, the feasible space is mapped to [0, 1] linearly.'
           'If `log`, the feasible space is mapped to [0, 1] logarithmically '
           'with formula: x -> log(x / min) / log(max / min). '
           'if `rlog`, the feasible space is mapped to [0, 1] "reverse" '
           'logarithmically, resulting in values close to `max_value` spread '
           'out more than the points near the `min_value`, with formula: '
           'x -> 1.0 - log((max + min - x) / min) / log (max / min). '
           '`min_value` must be positive if `scale` is not None. '
           'Also, it depends on the search algorithm to decide whether this '
           'information is used.'))


@symbolic.members(
    [
        ('min_value', pg_typing.Float(), 'Minimum value.'),
        ('max_value', pg_typing.Float(), 'Maximum value.'),
        float_scale_spec('scale')
    ],
    init_arg_list=[
        'min_value', 'max_value', 'scale', 'hints', 'location', 'name'],
    # TODO(daiyip): For backward compatibility.
    # Move this to additional keys later.
    serialization_key='pyglove.generators.geno.Float',
    additional_keys=['geno.Float'])
class Float(DecisionPoint):
  """Represents the genotype for a float-value genome.

  Example::

    # Create a float decision point within range [0.1, 1.0].
    decision_point = pg.geno.floatv(0.1, 1.0)

  See also: :func:`pyglove.geno.floatv`.
  """

  def _on_bound(self):
    """Custom logics to validate value."""
    super()._on_bound()
    if self.min_value > self.max_value:
      raise ValueError(
          f'Argument \'min_value\' ({self.min_value}) should be no greater '
          f'than \'max_value\' ({self.max_value}).')
    if self.scale in ['log', 'rlog'] and self.min_value <= 0:
      raise ValueError(
          f'\'min_value\' must be positive when `scale` is {self.scale!r}. '
          f'encountered: {self.min_value}.')

  @property
  def is_categorical(self) -> bool:
    """Returns True if current node is a categorical choice."""
    return False

  @property
  def is_subchoice(self) -> bool:
    """Returns True if current node is a subchoice of a multi-choice."""
    return False

  @property
  def is_numerical(self) -> bool:
    """Returns True if current node is numerical decision."""
    return True

  @property
  def is_custom_decision_point(self) -> bool:
    """Returns True if current node is a custom decision point."""
    return False

  @property
  def decision_points(self) -> List[DecisionPoint]:
    """Returns all decision points in their declaration order."""
    return [self]

  @property
  def space_size(self) -> int:
    """Returns the size of the search space. Use -1 for infinity."""
    return -1

  def _next_dna(self, dna: Optional['DNA'] = None) -> Optional['DNA']:
    """Returns the next DNA in the space represented by this spec.

    Args:
      dna: The DNA whose next will be returned. If None, `next_dna` will return
        the first DNA.

    Returns:
      The next DNA or None if there is no next DNA.
    """
    if dna is None:
      return DNA(self.min_value)
    # TODO(daiyip): Use hint for implementing stateful `next_dna`.
    raise NotImplementedError('`next_dna` is not supported on `Float` yet.')

  def _random_dna(
      self,
      random_generator: Union[random.Random, types.ModuleType],
      previous_dna: Optional[DNA]) -> DNA:
    """Returns a random DNA based on current spec."""
    del previous_dna
    return DNA(value=random_generator.uniform(self.min_value, self.max_value))

  def __len__(self) -> int:
    """Returns number of decision points in current space."""
    return 1

  def validate(self, dna: DNA) -> None:
    """Validate whether a DNA value conforms to this spec."""
    if not isinstance(dna.value, float):
      raise ValueError(
          f'Expect float value. Encountered: {dna.value}, '
          f'Location: {self.location.path}.')
    if dna.value < self.min_value:
      raise ValueError(
          f'DNA value should be no less than {self.min_value}. '
          f'Encountered {dna.value}, Location: {self.location.path}.')
    if dna.value > self.max_value:
      raise ValueError(
          f'DNA value should be no greater than {self.max_value}. '
          f'Encountered {dna.value}, Location: {self.location.path}.')
    if dna.children:
      raise ValueError(
          f'Float DNA should have no children. '
          f'Encountered: {dna.children!r}, Location: {self.location.path}.')

  def format(self,
             compact: bool = True,
             verbose: bool = True,
             root_indent: int = 0,
             show_id: bool = True,
             **kwargs):
    """Format this object."""
    if not compact:
      return super().format(compact, verbose, root_indent, **kwargs)
    if show_id:
      kvlist = [('id', str(self.id), '\'\'')]
    else:
      kvlist = []
    details = utils.kvlist_str(
        kvlist
        + [
            ('name', self.name, None),
            ('min_value', self.min_value, None),
            ('max_value', self.max_value, None),
            ('scale', self.scale, None),
            ('hints', self.hints, None),
        ]
    )
    return f'{self.__class__.__name__}({details})'


def floatv(
    min_value: float,
    max_value: float,
    scale: Optional[str] = None,
    hints: Any = None,
    location: utils.KeyPath = utils.KeyPath(),
    name: Optional[str] = None,
) -> Float:
  """Returns a Float specification.

  It creates the genotype for :func:`pyglove.floatv`.

  Example::

    spec = pg.geno.floatv(0.0, 1.0)

  Args:
    min_value: The lower bound of decision.
    max_value: The upper bound of decision.
    scale: An optional string as the scale of the range. Supported values
      are None, 'linear', 'log', and 'rlog'.
      If None, the feasible space is unscaled.
      If `linear`, the feasible space is mapped to [0, 1] linearly.
      If `log`, the feasible space is mapped to [0, 1] logarithmically with
        formula `x -> log(x / min) / log(max / min)`.
      If `rlog`, the feasible space is mapped to [0, 1] "reverse"
        logarithmically, resulting in values close to `max_value` spread
        out more than the points near the `min_value`, with formula:
        x -> 1.0 - log((max + min - x) / min) / log (max / min).
      `min_value` must be positive if `scale` is not None.
      Also, it depends on the search algorithm to decide whether this
      information is used or not.
    hints: An optional hint object.
    location: A ``pg.KeyPath`` object that indicates the location of the
      decision point.
    name: An optional global unique name for identifying this decision
      point.

  Returns:
    A ``pg.geno.Float`` object.

  See also:

    * :func:`pyglove.geno.constant`
    * :func:`pyglove.geno.space`
    * :func:`pyglove.geno.oneof`
    * :func:`pyglove.geno.manyof`
    * :func:`pyglove.geno.custom`
  """
  return Float(min_value, max_value, scale,
               hints=hints, location=location, name=name)



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyglove.core import utils
from pyglove.core.symbolic import unknown_symbols


class UnknownTypeTest(unittest.TestCase):

  def test_basics(self):
    t = unknown_symbols.UnknownType(name='__main__.ABC', args=[int, str])
    self.assertEqual(t.name, '__main__.ABC')
    self.assertEqual(t.args, [int, str])
    self.assertEqual(
        repr(t),
        '<unknown-type __main__.ABC>[<class \'int\'>, <class \'str\'>]'
    )
    self.assertEqual(
        t.to_json(),
        {
            '_type': 'type',
            'name': '__main__.ABC',
            'args': [
                {'_type': 'type', 'name': 'builtins.int'},
                {'_type': 'type', 'name': 'builtins.str'},
            ]
        }
    )
    self.assertEqual(utils.from_json(t.to_json(), convert_unknown=True), t)
    self.assertEqual(
        t(x=1, y=2),
        unknown_symbols.UnknownTypedObject(type_name='__main__.ABC', x=1, y=2)
    )


class UnknownFunctionTest(unittest.TestCase):

  def test_basics(self):
    t = unknown_symbols.UnknownFunction(name='__main__.foo')
    self.assertEqual(t.name, '__main__.foo')
    self.assertEqual(repr(t), '<unknown-function __main__.foo>')
    self.assertEqual(
        t.to_json(),
        {
            '_type': 'function',
            'name': '__main__.foo',
        }
    )
    self.assertEqual(utils.from_json(t.to_json(), convert_unknown=True), t)


class UnknownMethodTest(unittest.TestCase):

  def test_basics(self):
    t = unknown_symbols.UnknownMethod(name='__main__.ABC.bar')
    self.assertEqual(t.name, '__main__.ABC.bar')
    self.assertEqual(repr(t), '<unknown-method __main__.ABC.bar>')
    self.assertEqual(
        t.to_json(),
        {
            '_type': 'method',
            'name': '__main__.ABC.bar',
        }
    )
    self.assertEqual(utils.from_json(t.to_json(), convert_unknown=True), t)


class UnknownObjectTest(unittest.TestCase):

  def test_basics(self):
    v = unknown_symbols.UnknownTypedObject(type_name='__main__.ABC', x=1)
    self.assertEqual(v.type_name, '__main__.ABC')
    self.assertEqual(v.x, 1)
    self.assertEqual(repr(v), '<unknown-type __main__.ABC>(x=1)')
    self.assertEqual(
        str(v), '<unknown-type __main__.ABC>(\n  x = 1\n)')
    self.assertEqual(
        v.to_json(),
        {
            '_type': '__main__.ABC',
            'x': 1,
        }
    )
    self.assertEqual(utils.from_json(v.to_json(), convert_unknown=True), v)


if __name__ == '__main__':
  unittest.main()
